-- Legal Notice: (C)2006 Altera Corporation. All rights reserved.  Your
-- use of Altera Corporation's design tools, logic functions and other
-- software and tools, and its AMPP partner logic functions, and any
-- output files any of the foregoing (including device programming or
-- simulation files), and any associated documentation or information are
-- expressly subject to the terms and conditions of the Altera Program
-- License Subscription Agreement or other applicable license agreement,
-- including, without limitation, that your use is for the sole purpose
-- of programming logic devices manufactured by Altera and sold by Altera
-- or its authorized distributors.  Please refer to the applicable
-- agreement for further details.

LIBRARY IEEE, ALTERA;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE STD.textio.ALL;

-- A way for a CusP process to accept avalon slave accesses and do something
-- with them.  This is expected to be used as a configuration port so is
-- optimised for size rather than speed.

-- The accesses on the CusP side are controlled by two triggered sockets

-- CusP drives a 1 into the waitaccess socket to indicate that it is ready
-- for a new access.
--  * If no access is available then this stalls the CusP machine until
--    one becomes available.
--  * If an access is available then the control and write data are copied
--    into the CusP output sockets.
--  * The output sockets do not change except when triggered by waitaccess
--  * The hasaccess socket may be used to check whether waitaccess will
--    stall.

-- CusP drives a 1 into the finish socket to indicate that it has finished
-- with the last access.
--  * If the access was a read access then the read data must be provided
--    on the same cycle.

-- The CusP machine must not drive finish if there is no waiting access.

-- It may trigger waitaccess more than once for the same access, in this case
-- later triggers won't change any output sockets.

-- The CusP machine may drive waitaccess and finish on the same cycle,
-- this starts and finishes an access all in one go.

USE altera.ALT_CUSP151_PACKAGE.ALL;

entity alt_cusp151_avalon_mm_raw_slave is
  GENERIC (
        NAME            : STRING := "";
        WIDTH           : INTEGER := 16;
        ADDRESS_WIDTH   : INTEGER := 16;
        READ_USED       : INTEGER := 1;
        WRITE_USED      : INTEGER := 1;
        INTERRUPT_USED  : INTEGER := 1;
        CLOCKS_ARE_SYNC : INTEGER := 1
  );
  PORT (
        clock        : IN  STD_LOGIC;
        reset        : IN  STD_LOGIC;
        ena          : IN  STD_LOGIC := '1';
        stall        : OUT STD_LOGIC;

        waitaccess   : IN  STD_LOGIC := '0';
        waitaccess_en: IN  STD_LOGIC := '0';
        finish       : IN  STD_LOGIC := '0';
        finish_en    : IN  STD_LOGIC := '0';
        hasaccess    : OUT STD_LOGIC;

        isread       : OUT STD_LOGIC;
        address      : OUT STD_LOGIC_VECTOR( ADDRESS_WIDTH-1 DOWNTO 0);
        wdata        : OUT STD_LOGIC_VECTOR( WIDTH-1 DOWNTO 0);
        rdata        : IN  STD_LOGIC_VECTOR( WIDTH-1 DOWNTO 0) := (others=>'0');

        setirq       : IN  STD_LOGIC := '0';
        setirq_en    : IN  STD_LOGIC := '0';
        setirq_ena   : IN  STD_LOGIC := '1';
        
        av_address   : IN  STD_LOGIC_VECTOR( ADDRESS_WIDTH-1 DOWNTO 0) := (others=>'0');
        av_writedata : IN  STD_LOGIC_VECTOR( WIDTH-1 DOWNTO 0)  := (others=>'0');
        av_readdata  : OUT STD_LOGIC_VECTOR( WIDTH-1 DOWNTO 0);
        av_clock     : IN  STD_LOGIC := '0';
        av_write     : IN  STD_LOGIC := '0';
        av_chipselect: IN  STD_LOGIC := '0';
        av_reset     : IN  STD_LOGIC := '0';
        av_waitrequest : OUT  STD_LOGIC := '0';
        av_interrupt : OUT  STD_LOGIC := '0'
  );
END;


ARCHITECTURE rtl OF ALT_CUSP151_AVALON_MM_raw_SLAVE IS

    COMPONENT alt_cusp151_trigger_to_synced_pulse IS
    PORT (
        reset           : IN STD_LOGIC;
	    clk             : IN STD_LOGIC;
        sync_reset      : IN STD_LOGIC;
        sync_clk        : IN STD_LOGIC;
    	trigger         : IN STD_LOGIC;
        return_pulse    : IN STD_LOGIC;
        synced_pulse    : OUT STD_LOGIC
    );
    END COMPONENT;

  signal do_avalon_write : STD_LOGIC;
  signal do_avalon_read : STD_LOGIC;
  
  signal do_avalon_write_req : STD_LOGIC;
  signal do_avalon_read_req : STD_LOGIC;
  signal do_avalon_req : STD_LOGIC;
  signal do_avalon_ack : STD_LOGIC;
  signal do_avalon_ack_trigger : STD_LOGIC;
  
  signal av_address_int : STD_LOGIC_VECTOR( ADDRESS_WIDTH-1 DOWNTO 0);
  signal av_writedata_int : STD_LOGIC_VECTOR( WIDTH-1 DOWNTO 0);

  signal stall_int : STD_LOGIC;

  signal rdata_reg      : STD_LOGIC_VECTOR( WIDTH-1 DOWNTO 0);
  
  signal interrupt_int : STD_LOGIC;
  
  signal do_newaccess : STD_LOGIC;
  signal do_finish : STD_LOGIC;
  signal pending : STD_LOGIC;
  signal isread_int : STD_LOGIC;

BEGIN

 -- synopsys synthesis_off
watcher: process 
  variable rn : TTA_X_D_RN_T( 0 downto 0)  := 
      ( 
         new STRING'("rdata"), 
         others => null
      );
  variable rv : TTA_X_D_RV_T( 0 downto 0)  := 
      ( 
          new STD_LOGIC_VECTOR(rdata_reg'high downto rdata_reg'low), 
          others => null
      );
begin

  loop
    wait on rdata_reg;
    rv(0).all := rdata_reg;
    TTA_X_D_registerDump(NAME, rn, rv);
  end loop;
  --DEALLOCATE (rn(0));
  --DEALLOCATE (rv(0));
end process;
 -- synopsys synthesis_on

do_avalon_write <= av_chipselect and av_write; 
do_avalon_read <= av_chipselect and not av_write;
do_avalon_req <= do_avalon_write_req OR do_avalon_read_req;
av_waitrequest <= (av_write OR av_chipselect) AND NOT do_avalon_ack;         

sync_clocks: IF (CLOCKS_ARE_SYNC = 1) GENERATE
    do_avalon_write_req <= do_avalon_write;
    do_avalon_read_req <= do_avalon_read;
    av_address_int <= av_address;
    av_writedata_int <= av_writedata; 
    do_avalon_ack <= do_avalon_ack_trigger;
END GENERATE;
    
async_clocks: IF (CLOCKS_ARE_SYNC /= 1) GENERATE    
    
    write_req_generation: alt_cusp151_trigger_to_synced_pulse
      	PORT MAP (clk=>av_clock, reset=>av_reset, sync_clk=>clock, sync_reset=>reset, trigger=>do_avalon_write, return_pulse=>do_avalon_ack, synced_pulse=>do_avalon_write_req);        
    read_req_generation: alt_cusp151_trigger_to_synced_pulse
      	PORT MAP (clk=>av_clock, reset=>av_reset, sync_clk=>clock, sync_reset=>reset, trigger=>do_avalon_read, return_pulse=>do_avalon_ack, synced_pulse=>do_avalon_read_req);
        
    ack_generation: alt_cusp151_trigger_to_synced_pulse
		PORT MAP (clk=>clock, reset=>reset, sync_clk=>av_clock, sync_reset=>av_reset, trigger=>do_avalon_ack_trigger, return_pulse=>'1', synced_pulse=>do_avalon_ack);        
                  
    av_write_capture: process(av_clock, av_reset)
    begin
        IF av_reset = '1' THEN   
            av_address_int <= (others => '0'); 
            av_writedata_int <= (others => '0'); 
        ELSIF Rising_Edge(av_clock) THEN
            IF av_chipselect = '1' THEN
                av_address_int <= av_address;
            END IF;
            IF do_avalon_write = '1' THEN
                av_writedata_int <= av_writedata;
            END IF;
        END IF;
    end process;
END GENERATE;

do_newaccess <= waitaccess AND waitaccess_en;
do_finish <= finish AND finish_en;

stall_int <= do_newaccess AND NOT (do_avalon_req OR pending);
stall <= stall_int;

-- This state machine manages accesses in the CusP clock domain
pending_reg: process(clock, reset)
begin
    IF reset = '1' THEN   
        pending <= '0';
        do_avalon_ack_trigger <= '0';
        av_readdata <= (others => '0');
    ELSIF Rising_Edge(clock) THEN
        pending <= (do_avalon_req OR pending) AND NOT (do_avalon_ack_trigger);
        do_avalon_ack_trigger <= (do_avalon_req OR pending) AND do_finish AND ena;
        if (do_finish AND ena) = '1' then
       	   av_readdata <= rdata;
        end if;
    END IF;
end process;

-- TODO: Have different state machines if CusP process is write only or read only

    
cusp_capture: process(clock, reset)
begin
    if reset = '1' then
        isread <= '0';
        address <= (others => '0');
        wdata <= (others => '0');
        hasaccess <= '0';
    elsif Rising_Edge(clock) then
        IF (do_newaccess AND ena) = '1' THEN
            isread <= do_avalon_read;
    	       address <= av_address_int;
        	   wdata <= av_writedata_int;
        END IF;       
        IF (ena = '1') then
           hasaccess <= do_avalon_req OR (pending AND NOT do_avalon_ack_trigger);
        END IF; 
    end if;
end process;



av_interrupt <= interrupt_int;

no_irq_generate: IF INTERRUPT_USED = 0 GENERATE
  interrupt_int <= '0';
END GENERATE;

irq_generate: IF INTERRUPT_USED = 1 GENERATE

PROCESS(clock, reset)
BEGIN
    IF reset = '1' THEN
        interrupt_int <= '0';
    ELSIF Rising_Edge(clock) THEN
        IF (setirq_en AND setirq_ena) = '1' THEN
            interrupt_int <= setirq;
        END IF;
    END IF;
END PROCESS;

END GENERATE;

END ;

